use strict;
use warnings;

use Test::More tests => 7;
use_ok('EntityModel::Error');
use_ok('EntityModel::Hash');
use_ok('EntityModel::Array');
use_ok('EntityModel::Class::Accessor');
use_ok('EntityModel::Class::Accessor::Hash');
use_ok('EntityModel::Class::Accessor::Array');
use_ok('EntityModel::Class');

